"""ASAI can find the available actions, conditions, and other useful IAM information for AWS services.

This was written to make building detailed IAM policies with regional permissions boundaries less tedious.

## Features

* Search or list all AWS services
* List all regional or global AWS services
* Create a list of available IAM actions, wildcarding common prefixes
* Display all available IAM actions for a service
* Print all IAM information for a service.
"""

from dataclasses import dataclass, field
from typing import List

__author__ = """Shaun Martin"""
__email__ = "inhumantsar@protonmail.com"
__version__ = "0.1.3"


@dataclass(order=True)
class AWSService:
    """Representation of service information as received from policies.js."""

    name: str
    # matching JSON keys for the rest because i'm lazy
    StringPrefix: str
    Actions: List[str]
    HasResource: bool
    ARNFormat: str = field(default="")
    ARNRegex: str = field(default="")
    conditionKeys: List[str] = field(default_factory=lambda: [])


from ._backend import (
    get_actions,
    get_actions_with_wildcards,
    get_policies,
    get_service_by_prefix,
    get_services,
    search_services,
    service_is_regional,
)

__all__ = [
    "AWSService",
    "get_actions",
    "get_policies",
    "get_service_by_prefix",
    "get_services",
    "get_actions_with_wildcards",
    "search_services",
    "service_is_regional",
]



"""Main module."""

import json
import logging
import re
from typing import Dict, Generator, List, Tuple, Union

import requests
from fuzzywuzzy import fuzz

from asai import AWSService

log = logging.getLogger()

_POLICIES_URL = "https://awspolicygen.s3.amazonaws.com/js/policies.js"
# policies.js sets one big var, trim that var def and we have valid JSON
_POLICIES_PREFIX = "app.PolicyEditorConfig="
_CAMEL_REGEX = re.compile(r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)")


def get_policies() -> Dict:
    """Download all possible IAM actions, conditions, etc."""
    resp = requests.get(_POLICIES_URL)
    if resp.status_code != 200:
        log.error(f"Bad response from {_POLICIES_URL}")
        return

    resp_json = resp.content[len(_POLICIES_PREFIX) :]
    return json.loads(resp_json)


def get_service_by_prefix(prefix: str) -> Union[AWSService, None]:
    """Find a single service by its StringPrefix. Returns None if no matches are found."""
    try:
        return next(s for s in get_services() if s.StringPrefix == prefix)
    except StopIteration as e:
        return


def get_services() -> List[AWSService]:
    """Parse out all services from policies."""
    service_map = get_policies()["serviceMap"]
    return [AWSService(name=k, **v) for k, v in service_map.items()]


def get_actions() -> Dict[str, List[str]]:
    """Parse out all service prefixes and actions from policies."""
    services = get_services()
    return {v.StringPrefix: v.Actions for _, v in services.items()}


def get_actions_with_wildcards(service: AWSService) -> List[str]:
    """Return a service's list of actions, wildcarding any common prefixes."""
    actions = []
    for action in service.Actions:
        prefix = _get_action_prefix(action)
        wildcard = f"{prefix}*"
        if wildcard in actions:
            continue

        if _count_prefix_occurances(prefix, service.Actions) > 1:
            actions.append(wildcard)
        else:
            actions.append(action)
    return sorted(actions)


def search_services(
    search_term: str, min_confidence=70
) -> Generator[Tuple[int, AWSService], None, None]:
    """Use fuzzy matching to search the list of service names and prefixes."""
    search_term = search_term.lower()
    for service in get_services():
        name_words = service.name.lower().split()
        name_score = max([fuzz.ratio(i, search_term) for i in name_words])
        prefix_score = fuzz.ratio(service.StringPrefix, search_term)
        score = max(name_score, prefix_score)
        if score > min_confidence:
            yield (score, service)


def service_is_regional(svc: AWSService) -> bool:
    """Return true if a region field is present in the service's ARN format."""
    # region field can appear as ${region}, <region>, or region, in either title or lower case
    return "region" in svc.ARNFormat.lower()


def _get_action_prefix(action: str) -> str:
    """Split action by CamelCase, return prefix."""
    matches = re.findall(_CAMEL_REGEX, action)
    return matches[0]  # .group(0)


def _count_prefix_occurances(prefix: str, actions: List[str]) -> int:
    """Get the number of times a prefix occurs within a list of strings."""
    return len([i for i, x in enumerate(actions) if x.startswith(prefix)])



#!/usr/bin/env python

"""Tests for `asai` package."""

import pytest

from click.testing import CliRunner
from pathlib import Path
from os.path import join, dirname

from asai import _cli as cli


def _read(fname):
    """Read content of a file and return as a string."""
    return Path(join(dirname(__file__), fname)).read_text()


def test_cli_root():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    help_result = runner.invoke(cli.cli, ["--help"])
    assert help_result.exit_code == 0
    assert "--help  Show this message and exit." in help_result.output


def test_cli_services_list():
    types = ["all", "global", "no-arn", "regional"]
    expected = {i: _read(i) for i in types}
    runner = CliRunner()
    result = runner.invoke(cli.cli, args=["services"])
    assert result.exit_code == 0
    assert all([i in result.output for i in list(expected.keys())])
    for cmd, services in expected.items():
        result = runner.invoke(cli.cli, args=["services", cmd])
        assert result.exit_code == 0
        assert all([i in result.output for i in services])


def test_cli_services_search():
    runner = CliRunner()
    searches = {
        "auto": [
            "AWS Auto Scaling (autoscaling-plans)",
            "Amazon EC2 Auto Scaling (autoscaling)",
            "Application Auto Scaling (application-autoscaling)",
        ],
        "ec": [
            "Amazon EC2 Auto Scaling (autoscaling)",
            "Amazon EC2 (ec2)",
            "Amazon EC2 Instance Connect (ec2-instance-connect)",
            "Amazon Elastic Container Service (ecs)",
            "Amazon Elastic Container Registry (ecr)",
            "Amazon EC2 Image Builder (imagebuilder)",
        ],
        "container": [
            "Amazon Elastic Container Registry Public (ecr-public)",
            "Amazon Elastic Container Service (ecs)",
            "Amazon Elastic Container Registry (ecr)",
            "Amazon EMR on EKS (EMR Containers) (emr-containers)",
        ],
    }
    for term, results in searches.items():
        result = runner.invoke(cli.cli, args=["services", "search", term])
        assert result.exit_code == 0
        assert all([i in result.output for i in results])


def test_cli_single_service():
    types = ["actions", "actions-wildcard", "arn-format", "arn-regex", "info"]
    services = ["ec2", "s3", "dynamodb"]
    expected = {f"{i}_{j}": _read(f"{i}-{j}") for i in types for j in services}
    runner = CliRunner()
    result = runner.invoke(cli.cli, args=["service"])
    assert result.exit_code == 0
    assert all([i in result.output for i in types])
    for cmd, expected_results in expected.items():
        print(["service", *cmd.split("-")])
        result = runner.invoke(cli.cli, args=["service", *cmd.split("_")])
        assert result.exit_code == 0
        assert all([i in result.output for i in expected_results])



"""Console script for asai."""
import sys
from typing import List

import click

import asai


@click.group()
def cli():
    """ASAI can find the available actions, conditions, and other useful IAM information for AWS services."""
    pass


@cli.group("services")
def aws_services():
    """List or search for services by attribute."""
    pass


@aws_services.command("all")
def all_services():
    """List all service prefixes."""
    _echo_service_prefixes(asai.get_services())


@aws_services.command()
def no_arn():
    """List all services which don't have an ARN format."""
    services = asai.get_services()
    _echo_service_prefixes([s for s in services if not s.ARNFormat])


@aws_services.command("global")
def global_services():
    """List all services which aren't tied to specific regions."""
    services = asai.get_services()
    _echo_service_prefixes([s for s in services if not asai.service_is_regional(s)])


@aws_services.command()
def regional():
    """List all services which are tied to specific regions."""
    services = asai.get_services()
    _echo_service_prefixes([s for s in services if asai.service_is_regional(s)])


@aws_services.command()
@click.argument("search_term")
def search(search_term: str):
    """Search for a service by name or prefix."""
    results = asai.search_services(search_term)
    for _, svc in sorted(results, reverse=True, key=lambda x: x[0]):
        click.echo(f"{svc.name} ({svc.StringPrefix})")


@cli.group()
def service():
    """Get information on a single service."""
    pass


@service.command()
@click.argument("prefix")
def arn_format(prefix: str):
    """Display a service's ARN format description.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    click.echo(service.ARNFormat)


@service.command()
@click.argument("prefix")
def arn_regex(prefix: str):
    """Display a service's ARN Regex string.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    click.echo(service.ARNRegex)


@service.command()
@click.argument("prefix")
def actions(prefix: str):
    """List a service's actions.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    for action in sorted(service.Actions):
        click.echo(action)


@service.command()
@click.argument("prefix")
def actions_wildcard(prefix: str):  # noqa: D301
    """List a service's actions, wildcarding common prefixes.

    \b
        $ asai service actions-wildcard s3
        AbortMultipartUpload
        BypassGovernanceRetention
        Create*
        Delete*
        DescribeJob
        Get*
        List*
        ObjectOwnerOverrideToBucketOwner
        Put*
        Replicate*
        RestoreObject
        Update*

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)
    for action in asai.get_actions_with_wildcards(service):
        click.echo(action)


@service.command()
@click.argument("prefix")
def info(prefix: str):
    """Print all available service information.

    PREFIX is a service's short name. eg: ec2, ssm, or s3.
    """
    service = _get_service_by_prefix(prefix)

    click.echo(f"{service.name} ({service.StringPrefix})\n")
    click.echo(f"Has Resource: {service.HasResource}")
    click.echo("\nActions:")
    for action in sorted(service.Actions):
        click.echo(f"  - {action}")
    click.echo(f"\nARN Format: {service.ARNFormat}")
    click.echo(f"ARN Regex: {service.ARNRegex}")
    click.echo("\nCondition Keys:")
    for condition_key in service.conditionKeys:
        click.echo(f"  - {condition_key}")


def _echo_service_prefixes(services: List[asai.AWSService]):
    svc_ids = sorted(set([s.StringPrefix for s in services]))
    for svc in svc_ids:
        click.echo(svc)


def _get_service_by_prefix(prefix: str) -> asai.AWSService:
    service = asai.get_service_by_prefix(prefix)
    if not service:
        click.echo(f"{prefix} was not found in the service list.", err=True)
        sys.exit(1)
    return service


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover
